==> wyvern-2.1.2.tar.bz2: wyvern-2.1.2/util/suexec.c <==
/*****************************************************************************
 *
 * FILE:	suexec.c
 * DESCRIPTION:	wrapper module for suEXEC behavior for Wyvern
 * DATE:	Fri, Nov 29 2002
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2002-2004 졿Kouichi ABE (WALL), All rights reserved.
 * NOTE:	chown root:bin suexec && chmod 4111 suexec
 * REFERENCE:	SuExec of Apache and sbox (Put CGI Script in a Box)
 * LICENSE:
 *  Copyright (c) 2002-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: suexec.c,v 1.1.2.6 2004/12/19 13:55:32 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <stdarg.h>
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#if	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif	/* HAVE_SYS_PARAM_H */
#include <sys/resource.h>
#include <errno.h>
#include "suexec.h"


/*****************************************************************************
 *
 *	declaration of macros and structures
 *
 *****************************************************************************/
#define	EOL	'\0'
#define CR	'\015'
#define LF	'\012'

#define	ENVBUF	(64)


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	openLog(void);
static void	closeLog(void);
static void	logWriter(const char *, va_list);
static void	error(const char *, ...);
static void	die(const char *, ...);
static void	checkConsistency(void);
static void	checkTarget(const char *, struct stat *, char **, char **);
static void	cleanEnv(const char **, char ***);
#if	SET_LIMITS
static void	setLimits(void);
#endif	/* SET_LIMITS */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static char *	safe_env_list[] =
{
  "AUTH_TYPE",
  "CONTENT_LENGTH",
  "CONTENT_TYPE",
  "DATE_GMT",
  "DATE_LOCAL",
  "DOCUMENT_NAME",
  "DOCUMENT_ROOT",
  "DOCUMENT_URI",
  "GATEWAY_INTERFACE",
  "LAST_MODIFIED",
  "PATH_INFO",
  "PATH_TRANSLATED",
  "QUERY_STRING",
  "QUERY_STRING_UNESCAPED",
  "REMOTE_ADDR",
  "REMOTE_HOST",
  "REMOTE_IDENT",
  "REMOTE_PORT",
  "REMOTE_USER",
  "REQUEST_METHOD",
  "SCRIPT_NAME",
  "SERVER_ADMIN",
  "SERVER_NAME",
  "SERVER_PORT",
  "SERVER_PROTOCOL",
  "SERVER_ROOT",
  "SERVER_SOFTWARE",
  "TZ",
  NULL
};
static FILE *	log;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	openLog
 * DESCRIPTION:		open log file
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
openLog(void)
{
#ifdef	LOG_FILE
  if (strlen(LOG_FILE) > 0) {
    log = fopen(LOG_FILE, "a");
    if (log == NULL) {
      fprintf(stderr, "suexec failed to open log file: %s.", LOG_FILE);
      fprintf(stderr, "Falling back to stderr.\n");
      log = stderr;
    }
  }
  else {
    log = stderr;
  }
#else
  log = NULL;
#endif	/* LOG_FILE */
}

/*
 * FUNCTION NAME:	closeLog
 * DESCRIPTION:		close log file
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
closeLog(void)
{
  if (log != stderr) {
    fclose(log);
  }
}

/*
 * FUNCTION NAME:	logWriter
 * DESCRIPTION:		write message to log file
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
logWriter(fmt, ap)
	const char *	fmt;
	va_list		ap;
{
  time_t	now;
  char *	stamp;

  if (log == NULL) return;

  time(&now);
  stamp = ctime(&now);
  stamp[strlen(stamp) - 1] = EOL;

  fprintf(log, "[%s] suexec: ", stamp);
  vfprintf(log, fmt, ap);
  fflush(log);
}

/*
 * FUNCTION NAME:	error
 * DESCRIPTION:		write error message to log file
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
#if	__STDC__
error(const char *fmt, ...)
#else
error(fmt, va_alist)
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  logWriter(fmt, ap);
  va_end(ap);
}

/*
 * FUNCTION NAME:	die
 * DESCRIPTION:		write error message to log file and exit
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
#if	__STDC__
die(const char *fmt, ...)
#else
die(fmt, va_alist)
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  logWriter(fmt, ap);
  va_end(ap);
  exit(1);
}

/*
 * FUNCTION NAME:	checkConsistency
 * DESCRIPTION:		check the consitency of executable program
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
checkConsistency(void)
{
  uid_t			uid = getuid();
  gid_t			gid = getgid();
  struct passwd *	pwd;
  struct group *	grp;

  /* get our group and user ids */
  pwd = getpwuid(uid);
  if (pwd == NULL) {
    die("invalid uid: (%ld)\n", uid);
  }

  grp = getgrgid(gid);
  if (grp == NULL) {
    die("invalid gid: (%ld)\n", gid);
  }

  /* exit if either the uid or gid are zero */
  if (uid == 0) {
    die("cannot be run as root\n");
  }
  if (gid == 0) {
    die("cannot be run as root group\n");
  }

  if (strcmp(HTTPD_USER, pwd->pw_name) != 0) {
    die("invoked by %s, but must be run by %s\n", pwd->pw_name, HTTPD_USER);
  }
  if (strcmp(HTTPD_GROUP, grp->gr_name) != 0) {
    die("invoked by group %s, but must be run by group %s\n",
	grp->gr_name, HTTPD_GROUP);
  }
}

/*
 * FUNCTION NAME:	checkTarget
 * DESCRIPTION:		check the consitency of executable program
 * ARGUMENTS:		program name, status buffer, user name, group name
 * RETURN VALUE:	none
 */
static void
checkTarget(prog, file, uname, gname)
	const char *	prog;
	struct stat *	file;
	char **	uname;
	char **	gname;
{
  struct passwd *	pwd;
  struct group *	grp;

  if (stat(prog, file) == -1) {
    die("Stat failed: %s: %s\n", prog, strerror(errno));
  }
  if (!S_ISREG(file->st_mode)) {
    die("no valid executable script: %s\n", prog);
  }

  /* (1) script must not be SUID or SGID */
  if (file->st_mode & S_ISUID) {
    die("cannot run suid script (%s)\n", prog);
  }
  if (file->st_mode & S_ISGID) {
    die("cannot run sgid script (%s)\n", prog);
  }

  /* (2) script must be executable by other */
  if (!(file->st_mode & S_IXOTH)) {
    die("%s not executable\n", prog);
  }

  /* (3) script is not owned by HTTPD_USER or HTTPD_GROUP */
  pwd = getpwuid(file->st_uid);
  if (pwd == NULL) {
    die("%s is owned by an unknown user (%ld)\n", prog, file->st_uid);
  }
  if (strcmp(pwd->pw_name, HTTPD_USER) == 0) {
    die("%s is owned by the Web user (%s)\n", prog, HTTPD_USER);
  }
  *uname = strdup(pwd->pw_name);

  grp = getgrgid(file->st_gid);
  if (grp == NULL) {
    die("%s is owned by an unknown group (%ld)\n", prog, file->st_gid);
  }
  if (strcmp(grp->gr_name, HTTPD_GROUP) == 0) {
    die("%s is owned by the Web group (%s)\n", prog, HTTPD_GROUP);
  }
  *gname = strdup(grp->gr_name);

  /* (4) script is not writable by other */
  if (file->st_mode & S_IWOTH) {
    die("%s is wriable by other\n", prog);
  }

  /* (5) owner and group must not be less than UID_MIN, GID_MIN */
  if (file->st_uid < UID_MIN) {
    die("%s must not be owned by a UID less than %d\n", prog, UID_MIN);
  }
  if (file->st_gid < GID_MIN) {
    die("%s must not be owned by a GID less than %d\n", prog, GID_MIN);
  }
}

/*
 * FUNCTION NAME:	cleanEnv
 * DESCRIPTION:		make clean environment variables
 * ARGUMENTS:		environment, new environment
 * RETURN VALUE:	none
 */
static void
cleanEnv(environ, cleanenv)
	const char *	environ[];
	char ***	cleanenv;
{
  char *	path;
  char **	ep;
  int		cidx = 0;

  *cleanenv = (char **)calloc(ENVBUF, sizeof(char *));
  if (*cleanenv == NULL) {
    die("failed to allocate memory for environment\n");
  }

  if (asprintf(&path, "PATH=%s", SAFE_PATH) == -1) {
    die("failed to allocate memory for PATH");
  }
  (*cleanenv)[cidx] = path;
  cidx++;

  for (ep = (char **)environ; *ep && cidx < ENVBUF - 1; ep++) {
    if (strncmp(*ep, "HTTP_", 5) == 0 || strncmp(*ep, "SSL_", 4) == 0) {
      (*cleanenv)[cidx] = *ep;
      cidx++;
    }
    else {
      int	idx;
      for (idx = 0; safe_env_list[idx]; idx++) {
	if (strncmp(*ep, safe_env_list[idx], strlen(safe_env_list[idx])) == 0) {
	  (*cleanenv)[cidx] = *ep;
	  cidx++;
	  break;
	}
      }
    }
  }
  (*cleanenv)[cidx] = NULL;
}

#if	SET_LIMITS
/*
 * FUNCTION NAME:	setLimits
 * DESCRIPTION:		limit resource usage
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
setLimits(void)
{
#define SETLIMIT(a)	limit.rlim_cur = LIMIT_##a##_SOFT; \
			limit.rlim_max = LIMIT_##a##_HARD; \
			if (setrlimit(RLIMIT_##a,&limit)<0){ \
			  error("Couldn't set limit: %s",strerror(errno));}
  struct rlimit	limit;
#ifdef	RLIMIT_CPU
  SETLIMIT(CPU);
#endif	/* RLIMIT_CPU */
#ifdef RLIMIT_FSIZE
  SETLIMIT(FSIZE);
#endif
#ifdef RLIMIT_DATA
  SETLIMIT(DATA);
#endif
#ifdef RLIMIT_STACK
  SETLIMIT(STACK);
#endif
#ifdef RLIMIT_CORE
  SETLIMIT(CORE);
#endif
#ifdef RLIMIT_NOFILE
  SETLIMIT(NOFILE);
#endif
#ifdef RLIMIT_RSS
  SETLIMIT(RSS);
#endif
#ifdef RLIMIT_NPROC
  SETLIMIT(NPROC);
#endif
  setpriority(PRIO_PROCESS, 0, PRIORITY);
}
#endif	/* SET_LIMITS */

/*
 *	MAIN
 */
int
main(argc, argv, environ)
	int		argc;
	char *		argv[];
	const char *	environ[];
{
  char *	uname;	/* effective user name */
  char *	gname;	/* effective group name */
  char *	args[2] = { NULL, NULL };
  char *	prog;	/* executed program */
  struct stat	file;
  char **	cleanenv;

  if (argc == 1) {
    fprintf(stderr, "usage: %s program\n", argv[0]);
    exit(64);
  }

  /* set the script */
  prog = args[0] = argv[1];

  openLog();
  checkConsistency();
  checkTarget(prog, &file, &uname, &gname);
  cleanEnv(environ, &cleanenv);

  /* set to the owner's uid and gid */
  if (setgid(file.st_gid) < 0) {
    die("couldn't setgid() to %d (%s)\n", file.st_gid, strerror(errno));
  }
  if (setuid(file.st_uid) < 0) {
    die("couldn't setuid() to %d (%s)\n", file.st_uid, strerror(errno));
  }

#if	SET_LIMITS
  /* set hard and soft limits on resource usage */
  setLimits();
#endif	/* SET_LIMITS */

  error("executing %s for uid: %s(%d), gid: %s(%d)\n",
	prog, uname, file.st_uid, gname, file.st_gid);
  closeLog();

  /* execute the script */
  execve(prog, args, cleanenv);

  log = stderr;
  error("%s: exec failed (errno = %d: %s)\n", prog, errno, strerror(errno));

  return -1;
}

==> wyvern-2.2.10.tar.bz2: wyvern-2.2.10/util/suexec.c <==
/*****************************************************************************
 *
 * FILE:	suexec.c
 * DESCRIPTION:	Wyvern: wrapper program for suEXEC behavior
 * DATE:	Fri, Nov 29 2002
 * UPDATED:	Wed, Mar 31 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2004 졿Kouichi ABE (WALL), All rights reserved.
 * NOTE:	chown root:bin suexec && chmod 4111 suexec
 * REFERENCE:	SuExec of Apache and sbox (Put CGI Script in a Box)
 * LICENSE:
 *
 *  Copyright (c) 2002-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: suexec.c,v 1.4 2004/04/03 16:32:09 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <stdarg.h>
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#if	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif	/* HAVE_SYS_PARAM_H */
#include <sys/resource.h>
#include <errno.h>
#include "suexec.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	EOL	'\0'
#define CR	'\015'
#define LF	'\012'

#define	ENVBUF	(64)

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
#ifdef	LOG_FILE
static void	open_log_file(void);
static void	close_log_file(void);
static void	log_writer(const char *, va_list);
#endif	/* LOG_FILE */
static void	err(const char *, ...);
static void	die(const char *, ...);
static void	check_consistency(void);
static void	check_target(const char *, struct stat *, char **, char **);
static void	clean_environments(const char **, char ***);
#if	SET_LIMITS
static void	set_limits(void);
#endif	/* SET_LIMITS */

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static char *	safe_env_list[] =
{
  "AUTH_TYPE",
  "CONTENT_LENGTH",
  "CONTENT_TYPE",
  "DATE_GMT",
  "DATE_LOCAL",
  "DOCUMENT_NAME",
  "DOCUMENT_ROOT",
  "DOCUMENT_URI",
  "GATEWAY_INTERFACE",
  "LAST_MODIFIED",
  "PATH_INFO",
  "PATH_TRANSLATED",
  "QUERY_STRING",
  "QUERY_STRING_UNESCAPED",
  "REMOTE_ADDR",
  "REMOTE_HOST",
  "REMOTE_IDENT",
  "REMOTE_PORT",
  "REMOTE_USER",
  "REQUEST_METHOD",
  "SCRIPT_NAME",
  "SERVER_ADMIN",
  "SERVER_NAME",
  "SERVER_PORT",
  "SERVER_PROTOCOL",
  "SERVER_ROOT",
  "SERVER_SOFTWARE",
  "TZ",
  NULL
};
#ifdef	LOG_FILE
static FILE *	fp_log = NULL;
#endif	/* LOG_FILE */


/*****************************************************************************
 *
 *	Functions deinition
 *
 *****************************************************************************/

#ifdef	LOG_FILE
/*
 * FUNCTION NAME:	open_log_file
 * DESCRIPTION:		open log file
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
open_log_file(void)
{
  fp_log = fopen(LOG_FILE, "a");
  if (fp_log == NULL) {
    fprintf(stderr,
	    "suexec failed to open log file: %s."
	    "Falling back to stderr.\n", LOG_FILE);
    fp_log = stderr;
  }
}

/*
 * FUNCTION NAME:	close_log_file
 * DESCRIPTION:		close log file
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
close_log_file(void)
{
  if (fp_log != NULL && fp_log != stderr) {
    fclose(fp_log);
  }
}

/*
 * FUNCTION NAME:	log_writer
 * DESCRIPTION:		write message to log file
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
log_writer(fmt, ap)
	const char *	fmt;
	va_list		ap;
{
  time_t	now;
  char *	stamp;

  if (fp_log == NULL) return;

  time(&now);
  stamp = ctime(&now);
  stamp[strlen(stamp) - 1] = EOL;

  fprintf(fp_log, "[%s] suexec: ", stamp);
  vfprintf(fp_log, fmt, ap);
  fflush(fp_log);
}
#endif	/* LOG_FILE */

/*
 * FUNCTION NAME:	err
 * DESCRIPTION:		write error message to log file
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
#if	__STDC__
err(const char *fmt, ...)
#else
err(fmt, va_alist)
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
#ifdef	LOG_FILE
  log_writer(fmt, ap);
#endif	/* LOG_FILE */
  va_end(ap);
}

/*
 * FUNCTION NAME:	die
 * DESCRIPTION:		write error message to log file and exit
 * ARGUMENTS:		message
 * RETURN VALUE:	none
 */
static void
#if	__STDC__
die(const char *fmt, ...)
#else
die(fmt, va_alist)
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
#ifdef	LOG_FILE
  log_writer(fmt, ap);
#endif	/* LOG_FILE */
  va_end(ap);

  exit(1);
}

/*
 * FUNCTION NAME:	check_consistency
 * DESCRIPTION:		check the consitency of executable program
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
check_consistency(void)
{
  uid_t			uid = getuid();
  gid_t			gid = getgid();
  struct passwd *	pwd;
  struct group *	grp;

  /* get our group and user ids */
  pwd = getpwuid(uid);
  if (pwd == NULL) {
    die("invalid uid: (%ld)\n", uid);
  }

  grp = getgrgid(gid);
  if (grp == NULL) {
    die("invalid gid: (%ld)\n", gid);
  }

  /* exit if either the uid or gid are zero */
  if (uid == 0) {
    die("cannot be run as root\n");
  }
  if (gid == 0) {
    die("cannot be run as root group\n");
  }

  if (strcmp(HTTPD_USER, pwd->pw_name) != 0) {
    die("invoked by %s, but must be run by %s\n", pwd->pw_name, HTTPD_USER);
  }
  if (strcmp(HTTPD_GROUP, grp->gr_name) != 0) {
    die("invoked by group %s, but must be run by group %s\n",
	grp->gr_name, HTTPD_GROUP);
  }
}

/*
 * FUNCTION NAME:	check_target
 * DESCRIPTION:		check the consitency of executable program
 * ARGUMENTS:		program name, status buffer, user name, group name
 * RETURN VALUE:	none
 */
static void
check_target(prog, file, ret_uname, ret_gname)
	const char *	prog;
	struct stat *	file;
	char **		ret_uname;
	char **		ret_gname;
{
  struct passwd *	pwd;
  struct group *	grp;

  if (stat(prog, file) == -1) {
    die("Stat failed: %s: %s\n", prog, strerror(errno));
  }
  if (!S_ISREG(file->st_mode)) {
    die("no valid executable script: %s\n", prog);
  }

  /* (1) script must not be SUID or SGID */
  if (file->st_mode & S_ISUID) {
    die("cannot run suid script (%s)\n", prog);
  }
  if (file->st_mode & S_ISGID) {
    die("cannot run sgid script (%s)\n", prog);
  }

  /* (2) script must be executable by other */
  if (!(file->st_mode & S_IXOTH)) {
    die("%s not executable\n", prog);
  }

  /* (3) script is not owned by HTTPD_USER or HTTPD_GROUP */
  pwd = getpwuid(file->st_uid);
  if (pwd == NULL) {
    die("%s is owned by an unknown user (%ld)\n", prog, file->st_uid);
  }
  if (strcmp(pwd->pw_name, HTTPD_USER) == 0) {
    die("%s is owned by the Web user (%s)\n", prog, HTTPD_USER);
  }
  *ret_uname = strdup(pwd->pw_name);

  grp = getgrgid(file->st_gid);
  if (grp == NULL) {
    die("%s is owned by an unknown group (%ld)\n", prog, file->st_gid);
  }
  if (strcmp(grp->gr_name, HTTPD_GROUP) == 0) {
    die("%s is owned by the Web group (%s)\n", prog, HTTPD_GROUP);
  }
  *ret_gname = strdup(grp->gr_name);

  /* (4) script is not writable by other */
  if (file->st_mode & S_IWOTH) {
    die("%s is wriable by other\n", prog);
  }

  /* (5) owner and group must not be less than UID_MIN, GID_MIN */
  if (file->st_uid < UID_MIN) {
    die("%s must not be owned by a UID less than %d\n", prog, UID_MIN);
  }
  if (file->st_gid < GID_MIN) {
    die("%s must not be owned by a GID less than %d\n", prog, GID_MIN);
  }
}

/*
 * FUNCTION NAME:	clean_environments
 * DESCRIPTION:		make clean environment variables
 * ARGUMENTS:		environment, new environment
 * RETURN VALUE:	none
 */
static void
clean_environments(environ, ret_cleanenv)
	const char *	environ[];
	char ***	ret_cleanenv;
{
  char *	path;
  char **	ep;
  int		cidx = 0;

  *ret_cleanenv = (char **)calloc(ENVBUF, sizeof(char *));
  if (*ret_cleanenv == NULL) {
    die("failed to allocate memory for environment\n");
  }

  if (asprintf(&path, "PATH=%s", SAFE_PATH) == -1) {
    die("failed to allocate memory for PATH");
  }
  (*ret_cleanenv)[cidx] = path;
  cidx++;

  for (ep = (char **)environ; *ep && cidx < ENVBUF - 1; ep++) {
    if (strncmp(*ep, "HTTP_", 5) == 0 || strncmp(*ep, "SSL_", 4) == 0) {
      (*ret_cleanenv)[cidx] = *ep;
      cidx++;
    }
    else {
      int	idx;
      for (idx = 0; safe_env_list[idx]; idx++) {
	if (strncmp(*ep, safe_env_list[idx], strlen(safe_env_list[idx])) == 0) {
	  (*ret_cleanenv)[cidx] = *ep;
	  cidx++;
	  break;
	}
      }
    }
  }
  (*ret_cleanenv)[cidx] = NULL;
}

#if	SET_LIMITS
/*
 * FUNCTION NAME:	set_limits
 * DESCRIPTION:		limit resource usage
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
set_limits(void)
{
#define SETLIMIT(a)	limit.rlim_cur = LIMIT_##a##_SOFT; \
			limit.rlim_max = LIMIT_##a##_HARD; \
			if (setrlimit(RLIMIT_##a,&limit)<0){ \
			  err("Couldn't set limit: %s",strerror(errno));}
  struct rlimit	limit;
#ifdef	RLIMIT_CPU
  SETLIMIT(CPU);
#endif	/* RLIMIT_CPU */
#ifdef RLIMIT_FSIZE
  SETLIMIT(FSIZE);
#endif
#ifdef RLIMIT_DATA
  SETLIMIT(DATA);
#endif
#ifdef RLIMIT_STACK
  SETLIMIT(STACK);
#endif
#ifdef RLIMIT_CORE
  SETLIMIT(CORE);
#endif
#ifdef RLIMIT_NOFILE
  SETLIMIT(NOFILE);
#endif
#ifdef RLIMIT_RSS
  SETLIMIT(RSS);
#endif
#ifdef RLIMIT_NPROC
  SETLIMIT(NPROC);
#endif
  setpriority(PRIO_PROCESS, 0, PRIORITY);
}
#endif	/* SET_LIMITS */

/*
 *	MAIN
 */
int
main(argc, argv, environ)
	int		argc;
	char *		argv[];
	const char *	environ[];
{
  char *	uname;	/* effective user name */
  char *	gname;	/* effective group name */
  char *	args[2] = { NULL, NULL };
  char *	prog;	/* executed program */
  struct stat	file;
  char **	cleanenv;

  if (argc == 1) {
    fprintf(stderr, "usage: %s program\n", argv[0]);
    exit(64);
  }

  /* set the script */
  prog = args[0] = argv[1];

#ifdef	LOG_FILE
  open_log_file();
#endif	/* LOG_FILE */
  check_consistency();
  check_target(prog, &file, &uname, &gname);
  clean_environments(environ, &cleanenv);

  /* set to the owner's uid and gid */
  if (setgid(file.st_gid) < 0) {
    die("couldn't setgid() to %d (%s)\n", file.st_gid, strerror(errno));
  }
  if (setuid(file.st_uid) < 0) {
    die("couldn't setuid() to %d (%s)\n", file.st_uid, strerror(errno));
  }

#if	SET_LIMITS
  /* set hard and soft limits on resource usage */
  set_limits();
#endif	/* SET_LIMITS */

  err("executing %s for uid: %s(%d), gid: %s(%d)\n",
	prog, uname, file.st_uid, gname, file.st_gid);
#ifdef	LOG_FILE
  close_log_file();
#endif	/* LOG_FILE */

  /* execute the script */
  execve(prog, args, cleanenv);

#ifdef	LOG_FILE
  fp_log = stderr;
#endif	/* LOG_FILE */
  err("%s: exec failed (errno = %d: %s)\n", prog, errno, strerror(errno));

  return -1;
}
